#!/usr/bin/perl

my $version_string="0.27-jrm";
# some portions (c) 2003-2004 John McPherson

# sig2dot v0.27 (c) Darxus@ChaosReigns.com, released under the GPL
# Download from: http://www.chaosreigns.com/code/sig2dot/
#
# Parses output of "gpg --list-sigs" into a format
# suitable for rendering into a graph by graphviz 
# (http://www.research.att.com/sw/tools/graphviz/) like so:
#
# $ gpg --list-sigs --keyring ./phillylinux.gpg | ./sig2dot.pl > phillylinux.dot
# $ neato -Tps phillylinux.dot > phillylinux.ps
# $ convert phillylinux.ps phillylinux.jpg
#
# Commandline options:
#
# -b  
#   Black and white / do not colorize.
#
# -d <date>
#   Render graph as it appeared on <date> (ignores more recent
#   signatures).  Date must be in the format "YYYY-MM-DD".
#   Will also ignore keys that have since been revoked.


use strict;
use Getopt::Long;

# verbosity?
my $debug=0;

my $chartchar = "*";

#default options
my $renderdate="";
my $blackandwhite=0;
my ($help, $version);
if (!GetOptions('d' => \$renderdate, 'b' => \$blackandwhite, 'help' => \$help,
		'version' => \$version)
    or $help or $version) {
    print "$0 [options]\n";
    print "Version $version_string\n";
    if ($version) {exit}
    print "options:\n";
    print "  --help\tprint this help\n";
    print "  --version\tprint the version and exit\n";
    print "  -d\trender date\n";
    print "  -b\tuse black and white\n";
    exit 0;
}

if ($renderdate) { 
  print STDERR "Printing from date: $renderdate.\n";
  $renderdate =~ s/\D+//g;
}
if ($blackandwhite) 
{ 
  print STDERR "Black and White.\n";
}
# else { 
#  print STDERR "Color.\n";
#}

# hashes to store signing info
my %name;
my (%sigs, %sigcount, %signedby);
my @names;

my $line;
my $owner;
while ($line = <>)
{
  chomp $line;
  if ($line =~ m#([^ ]+)(?:\s\d+\s)? +([^ ]+) +([^ ]+) +([^<]+)#)
  {
    my ($type, $id, $date, $name) = ($1,$2,$3,$4);
    
    $date =~ tr/-//d;
    if ($type eq "pub" or $renderdate eq "" or $date <= $renderdate)
    {
      print STDERR "Using: $line\n" if ($debug);
      # strip trailing whitespace more cleanly:
      $name =~ s/\s+$//g;
  
      if ($type eq "pub")
      {
        $id = (split('/',$id))[1];
        $owner = $id; 
      } 
  
      $name{$id} = $name;
  
      # skip revoked keys 
      next if ($owner eq '[revoked]');
  
#      unless (defined @{$sigs{$owner}})
#      {
#        @{$sigs{$owner}} = ();
#      }
      if ($type eq "sig" and $id ne $owner and $name ne '[User id not found]')
      {
        push (@{$sigs{$owner}},$id);
        push (@{$signedby{$id}},$owner);
        push (@names,$id,$owner);
      }
    } else {
      print STDERR "Skipping due to date: $line\n" if ($debug);
    }
  } else {
    print STDERR "Skipping due to regex: $line\n" if ($debug);
  }
}

print "digraph \"debian-keyring\" {\noverlap=scale\nsplines=true\nsep=.1\n";

# get unique names out of the unsorted array
my %saw=();
@saw{@names} = ();
@names = keys %saw;

my $maxsigcount=0;
my $maxsignedbycount=0;
my %signedbycount;
my $maxratio=0;
for $owner (sort {$sigs{$a} <=> $sigs{$b}} keys %sigs)
{
  # get unique names for sigs
  %saw=();
  @saw{@{$sigs{$owner}}} = ();
  @{$sigs{$owner}} = keys %saw;

  # make sure this ID has signed at least one other key...
  if (exists $signedby{$owner}) {
      # get unique entries in list
      %saw=();
      @saw{@{$signedby{$owner}}} = ();
      @{$signedby{$owner}} = keys %saw;
  }
  $sigcount{$owner} = scalar(@{$sigs{$owner}});
  if ($sigcount{$owner} > $maxsigcount)
  {
    $maxsigcount = $sigcount{$owner};
  }

  if (exists ($signedby{$owner})) {
      $signedbycount{$owner} = scalar(@{$signedby{$owner}});
  } else {
      $signedbycount{$owner}=0;
  }
  if ($signedbycount{$owner} > $maxsignedbycount)
  {
    $maxsignedbycount = $signedbycount{$owner};
  }
  if ($signedbycount{$owner} / $sigcount{$owner} > $maxratio)
  {
    $maxratio = $signedbycount{$owner} / $sigcount{$owner};
  }
}
print "//$maxratio\n";

open (STATS,">stats.html");
print STATS "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">\n<html><head><title>Keyring Statistics</title></head><body><table border=1>\n";

for $owner (sort {$sigcount{$b} <=> $sigcount{$a}} keys %sigs)
{
  print STATS "<tr><td>$name{$owner}<td>$sigcount{$owner}<td><img src=\"/images/pipe0.jpg\" height=15 width=",$sigcount{$owner} * 20," alt=\"". $chartchar x $sigcount{$owner} ."\">\n";
}

print STATS "</table></body></html>\n";
close STATS;

print "node [style=filled]\n";
my ($red,$green,$blue,$hue,$saturation,$value);
for my $id (@names)
{
  if (!$blackandwhite)
  {
    $red = $sigcount{$id} / $maxsigcount;
    #$green = .25;
    if ($sigcount{$id})
    {
      $green = ($signedbycount{$id} / $sigcount{$id} / $maxratio * .75) * 2/3 + 1/3;
    } else {
      $green = 1/3;
    }
    $blue = ($signedbycount{$id} / $maxsignedbycount) * 2/3 + 1/3;
    ($hue,$saturation,$value) = rgb2hsv($red,$green,$blue);
    #print "//$red,$green,$blue\n";
    print "//$sigcount{$id} $signedbycount{$id} $red,$green,$blue\n";
    print "\"$id\" [fillcolor=\"$hue,$saturation,$value\",label=\"$name{$id}\"]\n";
  } else {
    print "\"$id\" [label=\"$name{$id}\"]\n";
  }
}
#print "node [style=solid]\n";

for $owner (sort keys %sigs)
{
 print "{ ";
  for my $id (@{$sigs{$owner}})
  {
    print "\"$id\" ";
  }
  print "} -> \"$owner\"\n";
}

print "}\n";

#  Converts rgb to hsv.  All numbers are within range 0 to 1
#  from http://twiki.org/cgi-bin/view/Codev/WebMap
sub rgb2hsv {
    my ($r, $g ,$b) = @_;
    my $max = maxof($r, maxof($g, $b));
    my $min = minof($r, minof($g, $b));
    my $v = $max;
    my ($s,$h);

    if ($max > 0.0) {
        $s = ($max - $min) / $max;
    } else {
        $s = 0;
    }
    if ($s > 0.0) {
        my ($rc, $gc, $bc, $diff);
            $diff = $max - $min;
        $rc = ($max - $r) / $diff;
        $gc = ($max - $g) / $diff;
        $bc = ($max - $b) / $diff;
        if ($r == $max) {
            $h = ($bc - $gc) / 6.0;
        } elsif ($g == $max) {
            $h = (2.0 + $rc - $bc) / 6.0;
        } else {
            $h = (4.0 + $gc - $rc) / 6.0;
        }
    } else {
       $h = 0.0;
    }
    if ($h < 0.0) {
       $h += 1.0;
    }
    return ($h, $s, $v);
}
sub maxof {
   my ($a, $b) = @_;

   return $a>$b?$a:$b;
}
sub minof {
   my ($a, $b) = @_;

   return $a<$b?$a:$b;
}


